/*
 * Kernel module for TQMP kernel mode media proxying.
 *
 * Shall work on post 4.18 kernels, where NAT API changed a lot (very late Deb9 and Deb10).
 *
 * To enable all functions, use command line params:
 * # modprobe do_stun_check=1 do_rtcp_check=1
 *
 * Copyright (c) 2000-2020 DEVERTO Ltd.
 */

#include <linux/module.h>
#include <net/ip.h>
#include <net/netfilter/nf_nat.h>
#include <net/netfilter/nf_conntrack_core.h>
#include <net/netfilter/nf_nat_l3proto.h>

MODULE_AUTHOR("Deverto Ltd <admin@deverto.com>");
MODULE_DESCRIPTION("Deverto TQMP kernel module");
MODULE_LICENSE("GPL"); // nf_ct_netns_get (and others) need this
MODULE_VERSION("1"); // version that understands the RTP flag

#define MODULENAME "nf_dvrt_tqmp"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// module parameters
// (Can be set from cmd line and set/get from /sys/module/nf_dvrt_tqmp/parameters/_paramName_.)
static short int do_stun_check = 0;
module_param(do_stun_check, short, S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH);
MODULE_PARM_DESC(do_stun_check, "0/1 disables/enables STUN packet NAT evasion (def: disabled)");
static short int do_rtcp_check = 0;
module_param(do_rtcp_check, short, S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH);
MODULE_PARM_DESC(do_rtcp_check, "0/1 disables/enables RTCP packet NAT evasion (def: disabled)");
static short int log_level = 0;
module_param(log_level, short, S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH);
MODULE_PARM_DESC(log_level, "0/1/2 off/NAT-evading/all packets (def: off)");


// these shall be in sync with TQMP
// conntrack mark set by TQMP:
// - 1st (most significant) bit: always 1 to see that it was created by TQMP
// - 17th bit: 1 for RTP stream
// - 2-16 bit: TQT network of one participant of the call
// - 18-32 bit: TQT network of the other participant of the call
#define NETIDMASK 0x7fff
#define REVERSESHIFT 16
#define MARKOFFSET 17
#define OWNPACKETCHKMASK 0x80000000
#define RTPPACKETCHKMASK 0x00008000

#define CREATE_IPV4_BYTES(ip) ip & 0xff, (ip >> 8) & 0xff, (ip >> 16) & 0xff, (ip >> 24) & 0xff

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Log the packet
static inline void log_packet(const char* prefix, const struct sk_buff* skb)
{
    const struct iphdr* iph;
    const struct udphdr* udph;
    const struct tcphdr* tcph;

    iph = ip_hdr(skb);

    if (iph->protocol == IPPROTO_UDP)
    {
	// Don't change this to udp_hdr(skb) here, it can be used only after packet hits transport layer!!!
	udph = (struct udphdr*)((void*)iph + ip_hdrlen(skb));

	pr_info("udp %s %d.%d.%d.%d:%d->%d.%d.%d.%d:%d\n",
		prefix,
		CREATE_IPV4_BYTES(iph->saddr),
		htons(udph->source),
		CREATE_IPV4_BYTES(iph->daddr),
		htons(udph->dest));
    }
    else if (iph->protocol == IPPROTO_TCP)
    {
	// Don't change this to tcp_hdr(skb) here, it can be only after packet hits transport layer!!!
	tcph = (struct tcphdr*)((void*)iph + ip_hdrlen(skb));

	pr_info("tcp %s %d.%d.%d.%d:%d->%d.%d.%d.%d:%d\n",
		prefix,
		CREATE_IPV4_BYTES(iph->saddr),
		htons(tcph->source),
		CREATE_IPV4_BYTES(iph->daddr),
		htons(tcph->dest));
    }
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// If packet is related to forwarded stream, return its conntrack and ctinfo.
static inline const struct nf_conn* is_tqmp_related(
	struct sk_buff* skb,
	enum ip_conntrack_info* ctinfo
	)
{
    // TQMP sends only IPv4 UDP and TCP to kernel mode.
    // For these a conntrack entry is created by TQMP with ct->mark 1st bit set

    const struct nf_conn* ct;
    const struct iphdr* iph;

    // TQMP forwards only IPv4 streams
    if (skb->protocol != htons(ETH_P_IP))
	return NULL;

    // TQMP forwards only UDP and TCP media streams
    iph = ip_hdr(skb);
    if (iph->protocol != IPPROTO_UDP && iph->protocol != IPPROTO_TCP)
	return NULL;

    // packet not in conntrack
    ct = nf_ct_get(skb, ctinfo);
    if (!ct)
	return NULL;

    // conntrack not created by TQMP (TQMP always sets 1st bit of mark)
    if (!(ct->mark & OWNPACKETCHKMASK))
	return NULL;

    return ct;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// If this STREAM is RTP
// TQMP sets 17th bit of conntrack mark for RTP streams
static inline bool is_rtp_stream(const struct nf_conn* ct)
{
    return ct->mark & RTPPACKETCHKMASK;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// If this is an RTCP packet
static inline bool is_rtcp(struct sk_buff* skb, const struct nf_conn* ct)
{
    const struct iphdr* iph;
    static const unsigned char buffer[4]; // to (optionally) store packet data chunk used for type check

    // MUST be used only for RTP/RTCP mixed stream, otherwise possible false RTCP identification takes place
    // (eg for UDPTL packets).
    if (!(do_rtcp_check && is_rtp_stream(ct)))
	return false;

    iph = ip_hdr(skb);

    if (iph->protocol == IPPROTO_UDP)
    {
	// Note: requested data will be copied to buffer only if:
	// - skb is nonlinear (aka paged) and
	// - needed data is in the paged area
	// (Otherwise we get a pointer to it in skb.)
	unsigned char* chunk = skb_header_pointer(
		skb,
		skb_network_offset(skb) + ip_hdrlen(skb) + sizeof(struct udphdr),
		sizeof(buffer),
		(void*) buffer
		);
	// data extraction failed, weird packet
	if (!chunk)
	    return false;

	// check RT(C)P version
	if ((chunk[0] >> 6) != 2)
	    return false;

	// see RFC5761 section 4
	// currently registered RTCP payloads and
	// RTCP payloads reserved for future use
	if ((chunk[1] >= 200 && chunk[1] <= 208) ||
		(chunk[1] >= 194 && chunk[1] <= 199) ||
		(chunk[1] >= 209 && chunk[1] <= 223))
	{
	    if (log_level)
	    {
		char logpref[50];
		snprintf(logpref, sizeof logpref, "rtcp packet payload:%d", chunk[1]);
		log_packet(logpref, skb);
	    }

	    return true;
	}
    }

    return false;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// If this is a STUN packet
static inline bool is_stun(struct sk_buff* skb)
{
    const struct iphdr* iph;
    size_t transporthdrsize;
    unsigned char* chunk = NULL;

    static const unsigned char stun_mc[] = {0x21, 0x12, 0xa4, 0x42}; // stun magic cookie (net byte order)
    static const short stun_mc_offset = 4; // stun magic cookie position in stun header

    static const unsigned char buffer[4]; // to (optionally) store packet data chunk used for type check

    // if not enabled in config
    if (!do_stun_check)
	return false;

    iph = ip_hdr(skb);

    if (iph->protocol == IPPROTO_UDP)
	transporthdrsize = sizeof(struct udphdr);
    else if (iph->protocol == IPPROTO_TCP)
	transporthdrsize = sizeof(struct tcphdr);
    else
	return false;

    // Note: requested data will be copied to buffer only if:
    // - skb is nonlinear (aka paged) and
    // - needed data is in the paged area
    // (Otherwise we get a pointer to it in skb.)
    chunk = skb_header_pointer(
	    skb,
	    skb_network_offset(skb) + ip_hdrlen(skb) + transporthdrsize + stun_mc_offset,
	    sizeof(buffer),
	    (void*) buffer
	    );

    // data extraction failed, weird packet
    if (!chunk)
	return false;

    if (!(memcmp(chunk, stun_mc, sizeof(stun_mc))))
    {
	if (log_level)
	    log_packet("stun packet", skb);

	return true;
    }

    return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Eliminate NAT by disassociating conntrack
static inline void remove_nat(struct sk_buff* skb, const struct nf_conn* ct)
{
    // This is achieved by unsetting the conntrack pointer that has been associated to this packet.
    // Without conntrack (and NAT info in it) routing will be done based on the original srcIP and dstIP,
    // and original srcIP is kept.
    nf_ct_set(skb, NULL, IP_CT_UNTRACKED);

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// copy conntrack mark's appropriate part to packet mark
static inline void set_mark(struct sk_buff* skb, const struct nf_conn* ct, enum ip_conntrack_info ctinfo)
{
    char* dir;

    if (CTINFO2DIR(ctinfo) == IP_CT_DIR_ORIGINAL)
    {
	skb->mark = ct->mark & NETIDMASK;
	dir = "orig"; // orig dir
    }
    else // IP_CT_DIR_REPLY
    {
	skb->mark = (ct->mark >> REVERSESHIFT) & NETIDMASK;
	dir = "reply"; // reply dir
    }

    // add a big number to packet marker (to avoid collision with used ones)
    skb->mark |= 1 << MARKOFFSET;

    if (log_level >= 2)
    {
	char logpref[100];
	snprintf(logpref, sizeof logpref, "packet marked (%s) mark:0x%04x (ctmark:0x%08x)", dir, skb->mark, ct->mark);
	log_packet(logpref, skb);
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// If packet has associated kernel auto-created conntrack with port SNAT
static inline const struct nf_conn* is_rogue_port_snat(
	struct sk_buff* skb,
	enum ip_conntrack_info* ctinfo,
	uint16_t* snatport
	)
{
    // Background:
    //
    // When in kernel forwarding, an (a-b;d-c) conntrack is created by TQMP (a-b is the orig, d-c is the reply dir,
    // b and c are the TQMP addresses).
    // When a STUN or RTCP packet is intercepted and sent to TQMP (eg. [a->b]), it will create and send
    // a reply packet [b->a].
    // A) This triggers an automatic conntrack creation attempt by kernel for (b-a;a-b) but because a-b is
    // already in use by the original (forwarder) conntrack, it will be changed to (b-a;a-b') where b' has
    // a new port (randomly  choosen by kernel). This activates a port SNAT, we must eliminate it.
    //
    // B) If at this state, the original (a-b;d-c) conntrack is deleted (by TQMP, because it is about to change stg)
    // and an [a->b] packet is received, it triggers auto conntrack creation by kernel for (a-b;b-a). This fails,
    // because the other rogue conntrack (described in A) still exists and occupies the reply addresses.
    // So, conntrack will be created as (a-b;b-a') with a' having random new port, activating port SNAT.

    const struct nf_conn* ct;
    const struct nf_conntrack_tuple* ct_orig_tuple;
    const struct nf_conntrack_tuple* ct_reply_tuple;
    const struct iphdr* iph;
    const struct udphdr* udph;
    const struct tcphdr* tcph;

    ct = nf_ct_get(skb, ctinfo);
    if (!ct)
	return NULL; // no conntrack

    // packets matching TQMP conntrack are not interesting here
    if (ct->mark & OWNPACKETCHKMASK)
	return NULL;

    // Only packets in original direction of the conntrack can have this kind of SNAT.
    if (CTINFO2DIR(*ctinfo) != IP_CT_DIR_ORIGINAL)
	return NULL;

    ct_reply_tuple = &ct->tuplehash[IP_CT_DIR_REPLY].tuple;
    ct_orig_tuple = &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple;
    iph = ip_hdr(skb);

    if (ct_orig_tuple->src.u3.ip != ct_reply_tuple->dst.u3.ip || // addr SNAT
	    ct_orig_tuple->dst.u3.ip != ct_reply_tuple->src.u3.ip) // addr DNAT
    {
	return NULL; // addr NAT can't be present
    }

    if (iph->protocol == IPPROTO_UDP)
    {
	// Don't change this to udp_hdr(skb) here, it can be used only after packet hits transport layer!!!
	udph = (struct udphdr*)((void*)iph + ip_hdrlen(skb));

	// check if port SNAT (and only that) is present
	// SNAT is stored in reply tuple dst part
	if (ct_orig_tuple->src.u.udp.port == ct_reply_tuple->dst.u.udp.port || //  no port SNAT
		ct_orig_tuple->dst.u.udp.port != ct_reply_tuple->src.u.udp.port) // port DNAT
	{
	    return NULL; // no port SNAT
	}

	*snatport = htons(ct_reply_tuple->dst.u.udp.port);
    }
    else if (iph->protocol == IPPROTO_TCP)
    {
	// Don't change this to tcp_hdr(skb) here, it can be only after packet hits transport layer!!!
	tcph = (struct tcphdr*)((void*)iph + ip_hdrlen(skb));

	// check if port SNAT (and only that) is present
	// SNAT is stored in reply tuple dst part
	if (ct_orig_tuple->src.u.tcp.port == ct_reply_tuple->dst.u.tcp.port || //  no port SNAT
		ct_orig_tuple->dst.u.tcp.port != ct_reply_tuple->src.u.tcp.port) // port DNAT
	{
	    return NULL; // no port SNAT
	}

	*snatport = htons(ct_reply_tuple->dst.u.tcp.port);
    }

    return ct;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Deverto packet manipulation steps on locally originated packets
static inline unsigned int orig_dvrt_process(void* priv, struct sk_buff* skb, const struct nf_hook_state* state)
{
    const struct nf_conn* ct = NULL;
    enum ip_conntrack_info ctinfo;
    uint16_t snatport = 0;

    if (!(ct = is_rogue_port_snat(skb, &ctinfo, &snatport)))
	return NF_ACCEPT;

    // we have port SNAT, get rid of it
    remove_nat(skb, ct);
    if (log_level)
    {
	char logpref[50];
	snprintf(logpref, sizeof logpref, "remove port SNAT %d", snatport);
	log_packet(logpref, skb);
    }

    return NF_ACCEPT; // all done, pass it on for other netfilter modules
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Deverto packet manipulation steps on incoming packets
static inline unsigned int inc_dvrt_process(void* priv, struct sk_buff* skb, const struct nf_hook_state* state)
{
    const struct nf_conn* ct;
    enum ip_conntrack_info ctinfo;
    uint16_t snatport = 0;

    // kernel created port SNAT, get rid of it
    if ((ct = is_rogue_port_snat(skb, &ctinfo, &snatport)))
    {
	remove_nat(skb, ct);
	if (log_level)
	{
	    char logpref[50];
	    snprintf(logpref, sizeof logpref, "remove port SNAT %d", snatport);
	    log_packet(logpref, skb);
	}

	// nothing else to do, not a TQMP forwarded packet
	return NF_ACCEPT;
    }

    if (!(ct = is_tqmp_related(skb, &ctinfo)))
	return NF_ACCEPT; // let it go

    if (is_stun(skb) || is_rtcp(skb, ct))
    {
	remove_nat(skb, ct); // send to local TQMP
	if (log_level)
	    log_packet("remove NAT", skb);
    }
    else
	set_mark(skb, ct, ctinfo); // set packet marker based on conntrack marker

    return NF_ACCEPT; // all done, pass it on for other netfilter modules
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// needed to start NAT, does nothing
// (Note: this function is actually only called for new conntrack entries and if a packet has no NAT info already set.
// This practically means that it is called very rarely.)
static inline unsigned int dummy_nat( void* priv, struct sk_buff* skb, const struct nf_hook_state* state)
{
    return NF_ACCEPT;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Deverto processing on originated packets
static const struct nf_hook_ops origprocessho =
{
    .hook = orig_dvrt_process,
    .hooknum = NF_INET_LOCAL_OUT, // output, processing locally originated packets
    .pf = NFPROTO_IPV4,
    .priority = NF_IP_PRI_CONNTRACK + 1 // right after conntrack (before any iptables added manip)
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Deverto processing on incoming packets
static const struct nf_hook_ops incprocessho =
{
    .hook = inc_dvrt_process,
    .hooknum = NF_INET_PRE_ROUTING, // prerouting, processing incomig packets
    .pf = NFPROTO_IPV4,
    .priority = NF_IP_PRI_CONNTRACK + 1 // right after conntrack
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// needed to start nat
static const struct nf_hook_ops natho =
{
    .hook = dummy_nat,
    .hooknum = NF_INET_PRE_ROUTING,
    .pf = NFPROTO_IPV4,
    .priority = NF_IP_PRI_NAT_DST + 1 // let all other nat function go before, we don't want to do anything
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static int __init load(void)
{
    int ret = 0;

    // start conntrack (register as conntrack user for IPV4)
    ret = nf_ct_netns_get(&init_net, NFPROTO_IPV4);
    if (ret)
    {
	pr_err("%s: module stop, cannot enable conntrack support, error: %d\n", MODULENAME, ret);
	return ret;
    }

    // start IPv4 NAT (start nf_nat_ipv4, aka nf_nat_l3proto_ipv4)
    // Note: without doing nf_nat_l3proto_ipv4_register_fn, nf_nat_ipv4 has no user
    // and does not process NAT info in conntrack
    // (till someone else registers as its user, eg. nftables by issuing some 'iptables -t nat ...' command).
    // Same true for IPv6.
    ret = nf_nat_l3proto_ipv4_register_fn(&init_net, &natho);
    if (ret)
    {
	nf_ct_netns_put(&init_net, NFPROTO_IPV4);
	pr_err("%s: module stop, cannot enable IPv4 NAT support, error: %d\n", MODULENAME, ret);
	return ret;
    }

    // start incoming packet manipulation
    ret = nf_register_net_hook(&init_net, &incprocessho);
    if (ret)
    {
	nf_nat_l3proto_ipv4_unregister_fn(&init_net, &natho);
	nf_ct_netns_put(&init_net, NFPROTO_IPV4);
	pr_err("%s: module stop, cannot start DVRT incoming packet manipulation, error: %d\n", MODULENAME, ret);
	return ret;
    }

    // start originated packet manipulation
    ret = nf_register_net_hook(&init_net, &origprocessho);
    if (ret)
    {
	nf_unregister_net_hook(&init_net, &incprocessho);
	nf_nat_l3proto_ipv4_unregister_fn(&init_net, &natho);
	nf_ct_netns_put(&init_net, NFPROTO_IPV4);
	pr_err("%s: module stop, cannot start DVRT originating packet manipulation, error: %d\n", MODULENAME, ret);
	return ret;
    }
    pr_info("%s: module started\n", MODULENAME);
    return ret;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void __exit remove(void)
{
    nf_nat_l3proto_ipv4_unregister_fn(&init_net, &natho);
    nf_unregister_net_hook(&init_net, &incprocessho);
    nf_unregister_net_hook(&init_net, &origprocessho);
    nf_ct_netns_put(&init_net, NFPROTO_IPV4);
    pr_info("%s: module stopped\n", MODULENAME);
}

module_init(load);
module_exit(remove);
